# -*- coding: utf-8 -*-
##############################################################################
#
#    This module is copyright (C) 2013 Numérigraphe SARL. All Rights Reserved.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
from collections import Iterable

from openerp.osv import orm, fields
from openerp.tools.translate import _
# The next 2 imports are only needed for a feature backported from trunk-wms
# TODOv8! remove, feature is included upstream
from openerp.osv import osv
from openerp.tools import DEFAULT_SERVER_DATETIME_FORMAT

from .exceptions import ExhaustiveInventoryException



class StockInventory(orm.Model):
    """Add locations to the inventories"""
    _inherit = 'stock.inventory'
    
    INVENTORY_STATE_SELECTION = [
                ('draft', 'Draft'),
                ('open', 'Open'),
                ('done', 'Done'),
                ('confirm', 'Confirmed'),
                ('cancel', 'Cancelled')
    ]
    _columns = {
        # TODOv8: don't add an extra state, use "confirm" = "In Progress" in v8
        'state': fields.selection(INVENTORY_STATE_SELECTION, 'State',
            readonly=True, select=True),
        # Make the inventory lines read-only in all states except "Open",
        # to ensure that no unwanted Location can be inserted
        'inventory_line_id': fields.one2many(
            'stock.inventory.line', 'inventory_id', 'Inventory lines',
            readonly=True, states={'open': [('readonly', False)]}),
        'location_ids': fields.many2many(
            'stock.location', 'stock_inventory_location_rel',
            'location_id', 'inventory_id',
            'Locations',
            readonly=True, states={'draft': [('readonly', False)]},
            help="This is the list of the Stock Locations that you want to "
                 "count the goods in.\n"
                 "Only these Locations can be entered in the "
                 "Inventory Lines.\n"
                 "If some of them have not been entered in the Inventory "
                 "Lines, OpenERP will warn you when you confirm the "
                 "Inventory."),
        'exhaustive': fields.boolean(
            'Exhaustive', readonly=True,
            states={'draft': [('readonly', False)]},
            help="""Check the box if you are conducting an exhaustive Inventory.
Leave the box unchecked if you are conducting a standard Inventory (partial inventory for example).
For an exhaustive Inventory:
 - the status "Draft" lets you define the list of Locations where goods must be counted
 - the status "Open" indicates that the list of Locations is definitive and you are now counting the goods. In that status, no Stock Moves can be recorded in/out of the Inventory's Locations
 - only the Inventory's Locations can be entered in the Inventory Lines
 - if some of the Inventory's Locations have not been entered in the Inventory Lines, OpenERP warns you when you confirm the Inventory
 - every good that is not in the Inventory Lines is considered lost, and gets moved out of the stock when you confirm the Inventory"""),
        }

    def action_open(self, cr, uid, ids, context=None):
        """Change the state of the inventory to 'open'"""
        # Refuse to open an exhaustive inventory without locations
        # TODOv7: isn't this unnecessary if location_id is required client-side? what harm would it do?
        for inventory in self.browse(cr, uid, ids, context=None):
            if inventory.exhaustive and not inventory.location_ids:
                raise ExhaustiveInventoryException(
                    _('Warning'), _('Location missing for this inventory.'))
        return self.write(cr, uid, ids, {'state': 'open'}, context=context)


    # TODOv8: remove this method, the feature is already done upstream
    def action_done(self, cr, uid, ids, context=None):
        """
        Don't allow to make an inventory with a date in the future.

        This makes sure no stock moves will be introduced between the
        moment you finish the inventory and the date of the Stock Moves.
        Backported from trunk-wms:
            revid:qdp-launchpad@openerp.com-20140317090656-o7lo22tzm8yuv3r8

        @raise osv.except_osv:
            We raise this exception on purpose instead of
            ExhaustiveInventoryException to ensure forward-compatibility
            with v8.
        """
        for inventory in self.browse(cr, uid, ids, context=None):
            if inventory.date > time.strftime(DEFAULT_SERVER_DATETIME_FORMAT):
                raise osv.except_osv(
                    _('Error!'),
                    _('It\'s impossible to confirm an inventory in the '
                      'future. Please change the inventory date to proceed '
                      'further.'))
        return super(StockInventory, self).action_done(cr, uid, ids,
                                                       context=context)

# TODOv7: set default location = Stock
# -> this is how trunk-wms does it (see also default location in inventory lines in v7):
#     def _default_stock_location(self, cr, uid, context=None):
#         try:
#             warehouse = self.pool.get('ir.model.data').get_object(cr, uid, 'stock', 'warehouse0')
#             return warehouse.lot_stock_id.id
#         except:
#             return False

    _defaults = {
        'state': 'draft',
        'exhaustive': False,
# TODOv7: set default location = Stock
#         'location_id': _default_stock_location,
        }

    def _check_location_free_from_inventories(self, cr, uid, ids):
        """
        Verify that no other Inventory is being conducted on the same locations

        Open Inventories are matched using the exact Location IDs,
        excluding children.
        """
        for inventory in self.browse(cr, uid, ids, context=None):
            if not inventory.exhaustive:
                # never block standard inventories
                continue
            location_ids = [location.id for location in inventory.location_ids]
            inv_ids = self.search(
                cr, uid, [('location_ids', 'in', location_ids),
                          ('id', '!=', inventory.id),
                          ('date', '=', inventory.date),
                          ('exhaustive', '=', True)])
            if inv_ids:
                # Quit as soon as one offending inventory is found
                return False
        return True

    _constraints = [
        (_check_location_free_from_inventories,
         'Other Physical inventories are being conducted using the same '
         'Locations.',
         ['location_ids', 'date', 'exhaustive'])
    ]

    def _get_locations_open_inventories(self, cr, uid, context=None):
        """IDs of location in open exhaustive inventories, with children"""
        open_inv_ids = self.search(cr, uid, [('state', '=', 'open'), ],
                                   context=context)
        location_ids = set()
        for open_inventory in self.browse(cr, uid, open_inv_ids,
                                          context=context):
            location_ids.update([location.id
                                 for location in open_inventory.location_ids])
        # Extend to the children Locations
        if location_ids:  # XXX probably works even otherwise
            location_ids = self.pool.get('stock.location').search(cr, uid,
                [('location_id', 'child_of', location_ids),
                 ('usage', '=', 'internal')],
                context=context)
        return location_ids

    def confirm_uninventoried_location_wizard(self, cr, uid, ids,
                                              context=None):
        """Open wizard to confirm empty locations on exhaustive inventories"""
        for inventory in self.browse(cr, uid, ids, context=context):
            if not inventory.exhaustive:
                return self.action_confirm(cr, uid, ids, context=context)
            else:
                # XXX: only open the wizard if it is needed
                context['active_ids'] = ids
                context['active_id'] = ids[0]
                return {
                    'type': 'ir.actions.act_window',
                    'view_type': 'form',
                    'view_mode': 'form',
                    'res_model': 'stock.inventory.uninventoried.locations',
                    'target': 'new',
                    'context': context,
                    'nodestroy': True,
                }

    # TODOv7: we probably don't need this anymore if we keep a single location per inventory like v8.
    # TODOv7: if we do need to keep this, we must rebase the monkey-patch on v7.0
    def _fill_location_lines(self, cr, uid, inventory_id, location_ids,
                             set_stock_zero, context=None):
        """
        Fill the inventory using all/only the inventory's locations.

        This is mostly a monkey patch on stock/stock_fill_inventory.py in v6.0

        @param self: The object pointer.
        @param cr: A database cursor
        @param uid: ID of the user currently logged in
        @param location_ids: the location ID or list of location IDs if we want more than one
        @param inventory_id: the inventory ID
        @param set_stock_zero: indicate if all the lines will be imported with zero quantity
        @param context: A standard dictionary
        @return:
        """
        inventory_line_obj = self.pool.get('stock.inventory.line')
        move_obj = self.pool.get('stock.move')
        uom_obj = self.pool.get('product.uom')
        res = []
        flag = False
        for location in location_ids:
            datas = {}
            move_ids = move_obj.search(cr, uid, ['|', ('location_dest_id', '=', location),
                                                      ('location_id', '=', location),
                                                      ('state', '=', 'done')], context=context)
            for move in move_obj.browse(cr, uid, move_ids, context=context):
                if move.location_dest_id.id == move.location_id.id:
                    continue
                lot_id = move.prodlot_id.id
                prod_id = move.product_id.id
                if move.location_dest_id.id == location:
                    qty = uom_obj._compute_qty(cr, uid, move.product_uom.id, move.product_qty, move.product_id.uom_id.id)
                else:
                    qty = -uom_obj._compute_qty(cr, uid, move.product_uom.id, move.product_qty, move.product_id.uom_id.id)
                if datas.get((prod_id, lot_id)):
                    qty = qty + datas[(prod_id, lot_id)]['product_qty']

                datas[(prod_id, lot_id)] = {'product_id': prod_id,
                                            'location_id': location,
                                            # Floating point sum could introduce some tiny rounding errors.
                                            # The uom are the same on input and output to use api for rounding.
                                            'product_qty': uom_obj._compute_qty_obj(cr, uid, move.product_id.uom_id, qty, move.product_id.uom_id),
                                            'product_uom': move.product_id.uom_id.id,
                                            'prod_lot_id': lot_id,
                                            'default_code': move.product_id.default_code,
                                            'prodlot_name': move.prodlot_id.name}
            if datas:
                flag = True
                res.append(datas)
        if not flag:
            raise ExhaustiveInventoryException(
                _('Warning'), _('No product in this location.'))

        stock_moves = []
        for stock_move in res:
            prod_lots = sorted(stock_move, key=lambda k: (stock_move[k]['default_code'], stock_move[k]['prodlot_name']))
            for prod_lot in prod_lots:
                stock_move_details = stock_move.get(prod_lot)

                if stock_move_details['product_qty'] == 0:
                    continue  # ignore product if stock equal 0

                stock_move_details.update({'inventory_id': inventory_id})

                if set_stock_zero:
                    stock_move_details.update({'product_qty': 0})

                domain = [(field, '=', stock_move_details[field])
                           for field in ['location_id',
                                         'product_id',
                                         'prod_lot_id']
                          ]
                # children_inventory_ids is present if stock_inventory_hierarchical_location module has been installed
                inventory_ids = context.get('children_inventory_ids', False)
                if inventory_ids:
                    domain.append(('inventory_id', 'child_of', inventory_ids))
                else:
                    domain.append(('inventory_id', '=', stock_move_details['inventory_id']))

                line_ids = inventory_line_obj.search(cr, uid, domain, context=context)
                if not line_ids:
                    stock_moves.append(stock_move_details)
        return stock_moves


class StockInventoryLine(orm.Model):
    """Only allow the Inventory's Locations"""

    _inherit = 'stock.inventory.line'

    # TODOv7: change the default location in the inventory lines to match that of the inventory
    # trunk-wms uses "context="{'default_location_id': location_id" on the field line_ids

    def onchange_location_id(self, cr, uid, ids, location_ids, exhaustive,
                             location_id, context=None):
        """Warn if the new is not in the location list for this inventory."""
        location_ids = location_ids[0][2]
        if not exhaustive or not location_ids:
            # Don't check if partial inventory
            return True

        # search children of location
        location_ids = self.pool.get('stock.location').search(
            cr, uid, [('location_id', 'child_of', location_ids)],
            context=context)
        if location_id not in location_ids:
            return {
                'value': {'location_id': False},
                'warning': {
                    'title': _('Warning: Wrong location'),
                    'message': _(
                         "You cannot record an Inventory Line for this "
                         "Location.\n"
                         "You must first add this Location to the list of "
                         "affected Locations on the Inventory form.")}
            }
        return True


class StockLocation(orm.Model):
    """Refuse changes during exhaustive Inventories"""
    _inherit = 'stock.location'
    _order = 'name'

    # TODOv7: why not put this in an ORM "_constraint" instead?
    def _check_inventory(self, cr, uid, ids, context=None):
        """Error if an exhaustive Inventory is being conducted here"""
        inv_obj = self.pool.get('stock.inventory')
        location_inventory_open_ids = inv_obj._get_locations_open_inventories(
            cr, uid, context=context)
        if not isinstance(ids, Iterable):
            ids = [ids]
        for inv_id in ids:
            if inv_id in location_inventory_open_ids:
                raise ExhaustiveInventoryException(
                    _('Error: Location locked down'),
                    _('A Physical Inventory is being conducted at this '
                      'location'))
        return True

    def write(self, cr, uid, ids, vals, context=None):
        """Refuse write if an inventory is being conducted"""
        self._check_inventory(cr, uid, ids, context=context)
        if not isinstance(ids, Iterable):
            ids = [ids]
        ids_to_check = ids
        # If changing the parent, no inventory must conducted there either
        if  vals.get('location_id'):
            ids_to_check.append(vals['location_id'])
        self._check_inventory(cr, uid, ids_to_check, context=context)
        return super(StockLocation, self).write(cr, uid, ids, vals,
                                                context=context)

    def create(self, cr, uid, vals, context=None):
        """Refuse create if an inventory is being conducted at the parent"""
        self._check_inventory(cr, uid, vals.get('location_id'),
                               context=context)
        return super(StockLocation, self).create(cr, uid, vals,
                                                 context=context)

    def unlink(self, cr, uid, ids, context=None):
        """Refuse unlink if an inventory is being conducted"""
        self._check_inventory(cr, uid, ids, context=context)
        return super(StockLocation, self).unlink(cr, uid, ids, context=context)


class StockMove(orm.Model):
    """Refuse Moves during exhaustive Inventories"""

    _inherit = 'stock.move'

    # TODOv7: adapt this to match trunk-wms
    def _check_open_inventory_location(self, cr, uid, ids, context=None):
        """
        Check that the locations are not locked by an open inventory

        Standard inventories are not checked.

        @raise ExhaustiveInventoryException: an error is raised if locations
            are locked, instead of returning False, in order to pass an
            extensive error message back to users.
        """
        message = ""
        inv_obj = self.pool.get('stock.inventory')
        locked_location_ids = inv_obj._get_locations_open_inventories(
            cr, uid, context=context)
        if not locked_location_ids:
            # Nothing to verify
            return True
        for move in self.browse(cr, uid, ids, context=context):
            if (move.location_id.usage != 'inventory'
                and move.location_dest_id.id in locked_location_ids):
                message += " - %s\n" % (move.location_dest_id.name)
            if (move.location_dest_id.usage != 'inventory'
                and move.location_id.id in locked_location_ids):
                message += " - %s\n" % (move.location_id.name)
        if message:
            raise ExhaustiveInventoryException(
                _('Error: Location locked down'),
                _('A Physical Inventory is being conducted at the following '
                  'location(s):\n%s') % message)
        return True

    _constraints = [
        (_check_open_inventory_location,
         "A Physical Inventory is being conducted at this location",
         ['location_id', 'location_dest_id']),
    ]
